use secrecy::{ExposeSecret, SecretString};
use url::Url;
use urlencoding::decode;

use crate::gitlib::GitUrl;
use crate::prelude::*;

/// The URL of a remote repository
#[derive(Clone, Debug)]
pub struct RemoteUrl {
    pub scheme: String,
    pub host: String,
    pub fullname: String,
    pub path: String,
    pub user: String,
    pub password: SecretString,
    pub(crate) original_url: String,
}

impl std::fmt::Display for RemoteUrl {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        write!(f, "{}", self.format_display())
    }
}

impl RemoteUrl {
    // Helper for the Display trait implementation and general formatting
    fn format_display(&self) -> String {
        if self.scheme == "file" {
            // For file URLs, just display the original path-like URL
            self.original_url.clone()
        } else {
            // Format for http/https/ssh etc.
            // If user is present, always show redacted password part, avoids exposing secret.
            let user_pass = if self.user.is_empty() {
                String::new()
            } else {
                format!("{}:{}@", self.user, "[REDACTED]")
            };
            format!(
                "{}://{}{}/{}",
                self.scheme, user_pass, self.host, self.fullname
            )
        }
    }

    pub fn parse(url_str: impl Into<String>) -> Option<Self> {
        let url_str = url_str.into();
        // Try parsing with url::Url first to handle file:// correctly
        if let Ok(parsed_url) = Url::parse(&url_str) {
            if parsed_url.scheme() == "file" {
                return Some(Self {
                    scheme: "file".to_string(),
                    host: parsed_url.host_str().unwrap_or_default().to_string(),
                    fullname: parsed_url.path().trim_start_matches('/').to_string(),
                    path: parsed_url.path().to_string(),
                    user: String::new(),
                    password: SecretString::default(),
                    original_url: url_str,
                });
            }
        }

        // Fallback to git_url_parse for other schemes (http, https, git, ssh)
        ::git_url_parse::GitUrl::parse(&url_str)
            .ok()
            .map(|parsed| Self {
                scheme: Url::parse(&url_str)
                    .map_or_else(|_| parsed.scheme.to_string(), |u| u.scheme().to_string()),
                host: parsed
                    .host
                    .map(|host| {
                        decode(&host).map_or_else(|_| host.clone(), std::borrow::Cow::into_owned)
                    })
                    .unwrap_or_default(),
                fullname: decode(&parsed.fullname)
                    .map_or_else(|_| parsed.fullname.clone(), std::borrow::Cow::into_owned),
                path: decode(&parsed.path)
                    .map_or_else(|_| parsed.path.clone(), std::borrow::Cow::into_owned),
                user: parsed
                    .user
                    .map(|u| decode(&u).map_or_else(|_| u.clone(), std::borrow::Cow::into_owned))
                    .unwrap_or_default(),
                password: parsed
                    .token
                    .map(|t| decode(&t).map_or_else(|_| t.clone(), std::borrow::Cow::into_owned))
                    .unwrap_or_default()
                    .into(),
                original_url: url_str,
            })
    }
}
impl TryFrom<GitUrl> for RemoteUrl {
    type Error = Error;

    fn try_from(value: GitUrl) -> Result<Self, Self::Error> {
        Self::parse(value.to_string()).ok_or_else(|| err!("Failed to parse GitUrl: {value}"))
    }
}
impl RemoteUrl {
    pub fn matches(&self, other: &Self) -> bool {
        debug!(?self, ?other);

        let scheme_match = self.scheme.eq_ignore_ascii_case(&other.scheme);
        let host_match = self.host.to_lowercase() == other.host.to_lowercase();
        let self_path_normalized = self.path.strip_suffix(".git").unwrap_or(&self.path);
        let other_path_normalized = other.path.strip_suffix(".git").unwrap_or(&other.path);
        let path_match = self_path_normalized.eq_ignore_ascii_case(other_path_normalized);
        let user_match = urlencoding::decode(&self.user)
            .unwrap_or_default()
            .eq_ignore_ascii_case(&urlencoding::decode(&other.user).unwrap_or_default());
        let password_match = self.password.expose_secret() == other.password.expose_secret();

        scheme_match && host_match && path_match && user_match && password_match
    }
}
