use std::path::Path;

use crate::gitlib::GitRepository;
use crate::prelude::*;
use crate::{
    core::{
        BranchName, ForgeAlias, ForgeConfig, ForgeDetails, GitDir, RemoteUrl, RepoAlias,
        RepoConfig, RepoPath, ServerRepoConfig,
    },
    git::{
        repository::open::{oreal::RealOpenRepository, OpenRepositoryLike},
        Generation,
    },
};

use instrument;
use secrecy::{ExposeSecret, SecretString};

/// The derived information about a repo, used to interact with it
#[derive(Builder, Clone, Debug, derive_more::Display, derive_with::With)]
#[display("gen-{}:{}:{}/{}", generation, forge.forge_type(), forge.forge_alias(), repo_alias )]
pub struct RepoDetails {
    pub generation: Generation,
    pub repo_alias: RepoAlias,
    pub repo_path: RepoPath,
    pub branch: BranchName,
    pub forge: ForgeDetails,
    pub repo_config: Option<RepoConfig>,
    pub gitdir: GitDir,
}
impl RepoDetails {
    #[must_use]
    pub fn new(
        generation: Generation,
        repo_alias: &RepoAlias,
        server_repo_config: &ServerRepoConfig,
        forge_alias: &ForgeAlias,
        forge_config: &ForgeConfig,
        gitdir: GitDir,
    ) -> Self {
        Self {
            generation,
            repo_alias: repo_alias.clone(),
            repo_path: server_repo_config.repo(),
            repo_config: server_repo_config.repo_config(),
            branch: server_repo_config.branch(),
            gitdir,
            forge: ForgeDetails::new(
                forge_alias.clone(),
                forge_config.forge_type(),
                forge_config.hostname(),
                forge_config.user(),
                forge_config.token(),
                forge_config.max_dev_commits(),
            ),
        }
    }
    pub(crate) fn origin(&self) -> secrecy::SecretString {
        let repo_details = self;
        let user = &repo_details.forge.user();
        let hostname = &repo_details.forge.hostname();

        let repo_path = &repo_details.repo_path;
        let expose_secret = repo_details.forge.token();

        let token = expose_secret.expose_secret();
        let origin = format!("https://{user}:{token}@{hostname}/{repo_path}.git");
        origin.into()
    }

    pub(crate) const fn gitdir(&self) -> &GitDir {
        &self.gitdir
    }

    #[cfg(test)]
    #[must_use]
    pub fn with_hostname(mut self, hostname: crate::core::Hostname) -> Self {
        let forge = self.forge;
        self.forge = forge.with_hostname(hostname);
        self
    }

    // url is a secret as it contains auth token
    pub(crate) fn url(&self) -> SecretString {
        let user = self.forge.user();
        let token = self.forge.token().expose_secret();
        let auth_delim = if token.is_empty() { "" } else { ":" };
        let hostname = self.forge.hostname();
        let repo_path = &self.repo_path;
        format!("https://{user}{auth_delim}{token}@{hostname}/{repo_path}.git").into()
    }

    #[allow(clippy::result_large_err)]
    pub(crate) fn open(&self) -> Result<impl OpenRepositoryLike, crate::Error> {
        let git_repo = GitRepository::open(self.gitdir().as_path())?;
        let repo = RealOpenRepository::builder()
            .inner(git_repo)
            .forge_details(self.forge.clone())
            .build();
        Ok(repo)
    }

    #[must_use]
    pub fn remote_url(&self) -> Option<RemoteUrl> {
        use secrecy::ExposeSecret;
        RemoteUrl::parse(self.url().expose_secret())
    }

    #[instrument]
    pub fn assert_remote_url(&self, found_via_default_backend: RemoteUrl) -> Result<()> {
        let Some(expected) = self.remote_url() else {
            debug!("No remote url to assert against");
            return Ok(());
        };
        let found_from_config = GitRepository::open(self.gitdir().as_path())
            .map_err(|e| err!("Failed to open repo: {e}"))?
            .find_default_remote_url()
            .map_err(|e| err!("Failed to read remote URL: {e}"))?;
        if !found_from_config.matches(&expected) {
            debug!(found = %found_from_config.original_url, expected = %expected.original_url, "URLs differ (exact match failed)");
            self.write_remote_url(&expected)?;
        }
        Ok(())
    }

    #[instrument]
    pub fn write_remote_url(&self, url: &RemoteUrl) -> Result<()> {
        let path_part = {
            let base = if url.fullname.starts_with('/') && url.fullname.contains('/') {
                &url.fullname
            } else {
                &url.path
            };
            let slashed = if base.starts_with('/') {
                base.to_string()
            } else {
                format!("/{base}")
            };
            if Path::new(&slashed)
                .extension()
                .is_some_and(|ext| ext.eq_ignore_ascii_case("git"))
            {
                slashed
            } else {
                format!("{slashed}.git")
            }
        };
        let url_to_write = format!(
            "{}://{}:{}@{}{}",
            url.scheme,
            url.user,
            url.password.expose_secret(),
            url.host,
            path_part
        );

        // Open the repository using the git2 backend specifically for config writing.
        let git_repo = GitRepository::open(self.gitdir().as_path())?;

        // Get the repository's local configuration file (.git/config)
        let mut config = git_repo
            .lock()?
            .config()
            .map_err(|e| err!("open git config: {e}"))?
            .open_level(git2::ConfigLevel::Local)
            .map_err(|e| err!("open .git/config: {e}"))?;

        config
            .set_str("remote.origin.url", &url_to_write)
            .map_err(|e| err!("set remote origin url: {e}"))?;

        debug!(url = %url, "Successfully updated remote.origin.url using git2");
        Ok(())
    }
}
