use crate::gitlib::GitRepository;
use crate::prelude::*;
use crate::{
    core::git::repository::open::oreal::NonBareRealOpenRepository,
    git::{
        repository::open::{oreal::RealOpenRepository, OpenRepositoryLike},
        RepoDetails,
    },
};
use color_eyre::eyre::WrapErr;

use super::open::NonBareRepositoryLike;

#[allow(clippy::module_name_repetitions)]
#[mockall::automock]
pub trait RepositoryFactory: std::fmt::Debug + Sync + Send {
    fn duplicate(&self) -> Box<dyn RepositoryFactory>;

    /// Opens the repository.
    ///
    /// # Errors
    ///
    /// Will return an `Err` if the repository can't be opened.
    fn open(&self, repo_details: &RepoDetails) -> Result<Box<dyn OpenRepositoryLike>>;

    /// Clones a worktree for a git repository from the remote server.
    ///
    /// # Errors
    ///
    /// Will return an `Err` if there are any network connectivity issues
    /// connecting with the server.
    fn git_clone(&self, repo_details: &RepoDetails) -> Result<Box<dyn NonBareRepositoryLike>>;

    /// Clones a bare git repository from the remote server.
    ///
    /// # Errors
    ///
    /// Will return an `Err` if there are any network connectivity issues
    /// connecting with the server.
    fn git_clone_bare(&self, repo_details: &RepoDetails) -> Result<Box<dyn OpenRepositoryLike>>;
}

#[must_use]
pub fn real() -> Box<dyn RepositoryFactory> {
    Box::new(RealRepositoryFactory)
}

#[cfg(test)]
#[must_use]
pub fn mock() -> MockRepositoryFactory {
    MockRepositoryFactory::new()
}

#[derive(Debug, Clone)]
struct RealRepositoryFactory;

impl RepositoryFactory for RealRepositoryFactory {
    fn open(&self, repo_details: &RepoDetails) -> Result<Box<dyn OpenRepositoryLike>> {
        let repo = repo_details.open().wrap_err("Failed to open repository")?;

        let found = repo
            .find_default_remote()
            .wrap_err("Failed to find default remote")?;

        repo_details
            .assert_remote_url(found)
            .wrap_err("Remote URL validation failed")?;

        Ok(Box::new(repo))
    }

    fn git_clone(&self, repo_details: &RepoDetails) -> Result<Box<dyn NonBareRepositoryLike>> {
        let repo = NonBareRealOpenRepository::new(
            RealOpenRepository::builder()
                .inner(GitRepository::clone(repo_details)?)
                .forge_details(repo_details.forge.clone())
                .build(),
        );
        Ok(Box::new(repo))
    }

    #[instrument]
    fn git_clone_bare(&self, repo_details: &RepoDetails) -> Result<Box<dyn OpenRepositoryLike>> {
        let clone_bare = GitRepository::clone_bare(repo_details)?;
        let repo = RealOpenRepository::builder()
            .inner(clone_bare)
            .forge_details(repo_details.forge.clone())
            .build();
        Ok(Box::new(repo))
    }

    fn duplicate(&self) -> Box<dyn RepositoryFactory> {
        Box::new(self.clone())
    }
}
