use crate::prelude::*;
use crate::{
    core::RemoteUrl,
    git::{
        repository::open::OpenRepositoryLike, validation::remotes::validate_default_remote,
        RepoDetails,
    },
    gitlib::{GitCloneError, GitOpenError},
};

use info;

pub mod factory;
pub mod open;

#[cfg(test)]
mod test;

#[allow(clippy::module_name_repetitions)]
pub use factory::RepositoryFactory;

#[cfg(test)]
mod tests;

// #[cfg(test)]
// #[derive(Clone, Debug)]
// #[allow(clippy::large_enum_variant)]
// pub enum Repository {
//     Real,
//     Test(test::TestRepository),
// }

#[cfg(test)]
pub const fn test(
    fs: kxio::fs::FileSystem,
    forge_details: crate::core::ForgeDetails,
) -> test::TestRepository {
    test::TestRepository::new(fs, vec![], vec![], forge_details)
}

/// Opens a repository, cloning if necessary
#[instrument(skip_all)]
#[cfg(not(tarpaulin_include))] // requires network access to either clone new and/or fetch.
pub fn open(
    repository_factory: &dyn factory::RepositoryFactory,
    repo_details: &RepoDetails,
) -> Result<Box<dyn OpenRepositoryLike>> {
    let open_repository = if repo_details.gitdir.exists() {
        info!("Local copy found - opening...");
        repository_factory
            .open(repo_details)
            .map_err(|e| Error::Validation(e.to_string()))?
    } else {
        info!("Local copy not found - cloning...");
        repository_factory
            .git_clone_bare(repo_details)
            .map_err(Error::from)?
    };
    validate_default_remote(&*open_repository, repo_details)
        .map_err(|e| Error::Validation(s!(e)))?;
    Ok(open_repository)
}

#[cfg(test)]
#[allow(clippy::module_name_repetitions)]
pub trait RepositoryLike {
    /// Opens the repository.
    ///
    /// # Errors
    ///
    /// Will return an `Err` if the repository can't be opened.
    fn open(&self, gitdir: &crate::core::GitDir) -> Result<open::OpenRepository>;

    // /// Clones the git repository from the remote server.
    // ///
    // /// # Errors
    // ///
    // /// Will return an `Err` if there are any network connectivity issues
    // /// connecting with the server.
    // #[cfg(test)]
    // fn git_clone(&self, repo_details: &RepoDetails) -> Result<open::OpenRepository>;
}

pub type Result<T> = core::result::Result<T, Error>;

#[derive(Debug, thiserror::Error)]
pub enum Error {
    #[error("eyre error: {0}")]
    Eyre(#[from] color_eyre::eyre::Error),
    #[error("kxiofs: {0}")]
    KxioFs(#[from] kxio::fs::Error),

    #[error("validation: {0}")]
    Validation(String),

    #[error("git clone: {0}")]
    Clone(#[from] GitCloneError),

    #[error("open: {0}")]
    Open(#[from] Box<GitOpenError>), // Box the large error variant

    #[error("git fetch: {0}")]
    Fetch(String),

    #[error("MismatchDefaultFetchRemote(found: {found:?}, expected: {expected:?})")]
    MismatchDefaultFetchRemote {
        found: Box<RemoteUrl>,
        expected: Box<RemoteUrl>,
    },
}
