use crate::gitlib::Force;
use crate::prelude::*;
use crate::{
    core::{BranchName, ForgeDetails, Hostname, RemoteUrl, RepoPath},
    git::{self},
    gitlib::{self, GitRepository},
};

use derive_more::Constructor;

use std::{ops::Deref, path::Path};

use super::OpenRepositoryLike;

#[derive(Builder, Debug)]
pub struct RealOpenRepository {
    inner: GitRepository,
    forge_details: ForgeDetails,
}
impl RealOpenRepository {}
impl super::OpenRepositoryLike for RealOpenRepository {
    fn remote_branches(&self) -> Result<Vec<BranchName>> {
        self.inner.remote_branches()
    }

    #[instrument]
    fn find_default_remote(&self) -> Result<RemoteUrl> {
        self.inner.find_default_remote_url()
    }

    #[instrument(skip_all)]
    #[cfg(not(tarpaulin_include))] // would require writing to external service
    fn fetch(&self) -> Result<()> {
        info!("Fetching via gitlib");
        self.inner
            .fetch()
            .map_err(|e| err!("Git fetch failed: {e}"))?;
        info!("Fetch completed");
        Ok(())
    }

    #[cfg(not(tarpaulin_include))] // would require writing to external service
    #[instrument(skip_all)]
    fn push(
        &self,
        _repo_details: &git::RepoDetails,
        branch_name: &BranchName,
        to_commit: &git::GitRef,
        force: &gitlib::Force,
    ) -> Result<()> {
        info!(
            "Pushing via gitlib: branch={}, commit={}, force={}",
            branch_name, to_commit, force
        );

        match self.inner.push(to_commit, branch_name, force.clone()) {
            Ok(()) => {
                info!("Push completed successfully");
                Ok(())
            }
            Err(e) => {
                error!("Push failed: {}", e);
                Err(err!("Git push failed: {e}"))
            }
        }
    }

    fn commit_log(
        &self,
        branch_name: &BranchName,
        find_commits: &[git::Commit],
    ) -> Result<Vec<git::Commit>> {
        let limit: usize = if find_commits.is_empty() {
            1
        } else {
            self.forge_details
                .max_dev_commits()
                .map_or(25, |commit_count| commit_count.clone().peel() as usize)
        };
        self.inner.commit_log(branch_name, limit, find_commits)
    }

    #[instrument(skip_all, fields(%branch_name, ?file_name))]
    fn read_file(&self, branch_name: &BranchName, file_name: &Path) -> Result<String> {
        self.inner.read_file(branch_name, file_name)
    }

    #[instrument(skip(self), fields(%branch_name))]
    fn switch_branch(&self, branch_name: &str) -> Result<()> {
        self.inner.switch_branch(branch_name)
    }

    #[instrument(skip(self), fields(%new_base, %old_base, %branch))]
    fn rebase_onto(&self, new_base: &str, old_base: &str, branch: &str) -> Result<()> {
        self.inner.rebase_onto(new_base, old_base, branch)
    }
}

#[derive(Debug, Constructor)]
pub struct NonBareRealOpenRepository {
    inner: RealOpenRepository,
}
impl Deref for NonBareRealOpenRepository {
    type Target = RealOpenRepository;

    fn deref(&self) -> &Self::Target {
        &self.inner
    }
}

impl OpenRepositoryLike for NonBareRealOpenRepository {
    fn remote_branches(&self) -> Result<Vec<BranchName>> {
        self.deref().remote_branches()
    }

    fn find_default_remote(&self) -> Result<RemoteUrl> {
        self.deref().find_default_remote()
    }

    fn fetch(&self) -> Result<()> {
        self.deref().fetch()
    }

    fn push(
        &self,
        repo_details: &git::RepoDetails,
        branch_name: &BranchName,
        to_commit: &git::GitRef,
        force: &Force,
    ) -> Result<()> {
        self.deref()
            .push(repo_details, branch_name, to_commit, force)
    }

    fn commit_log(
        &self,
        branch_name: &BranchName,
        find_commits: &[git::Commit],
    ) -> Result<Vec<git::Commit>> {
        self.deref().commit_log(branch_name, find_commits)
    }

    fn read_file(&self, branch_name: &BranchName, file_name: &Path) -> Result<String> {
        self.deref().read_file(branch_name, file_name)
    }

    fn switch_branch(&self, branch_name: &str) -> Result<()> {
        self.deref().switch_branch(branch_name)
    }

    fn rebase_onto(&self, target_branch: &str, old_base: &str, branch: &str) -> Result<()> {
        self.deref().rebase_onto(target_branch, old_base, branch)
    }
}
impl super::NonBareRepositoryLike for NonBareRealOpenRepository {
    fn checkout(&self, branch_name: &BranchName) -> Result<()> {
        info!("Checking out branch via gitlib");
        self.inner
            .switch_branch(&branch_name.to_string())
            .map_err(|e| err!("Git checkout failed: {e}"))?;
        info!("Checkout completed");
        Ok(())
    }

    fn rebase(&self, target_branch_name: &BranchName) -> Result<()> {
        info!("Rebasing via gitlib");
        self.inner
            .rebase_onto(
                &target_branch_name.to_string(),
                "origin/main", // old base - assuming main is the integration branch
                "HEAD",        // branch to rebase - using current HEAD
            )
            .map_err(|e| err!("Git rebase failed: {e}"))?;
        info!("Rebase completed");
        Ok(())
    }
}

impl From<&RemoteUrl> for git::GitRemote {
    fn from(url: &RemoteUrl) -> Self {
        let path = url.path.clone();
        let path = path.strip_prefix('/').map_or(path.as_str(), |path| path);
        let path = path.strip_suffix(".git").map_or(path, |path| path);
        Self::new(Hostname::new(url.host.clone()), RepoPath::new(s!(path)))
    }
}
