use crate::prelude::*;
use crate::{
    core::RemoteUrl,
    git::{self, repository::open::OpenRepositoryLike},
};

#[derive(Debug, thiserror::Error)]
pub enum GitRemoteValidationError {
    #[error("No default push remote configured in repository")]
    NoDefaultPushRemote,
    #[error("No default fetch remote configured in repository")]
    NoDefaultRemote,
    #[error("Push remote mismatch: found {found}, expected {expected}")]
    RemoteMismatch {
        found: RemoteUrl,
        expected: RemoteUrl,
    },
    #[error("Fetch remote mismatch: found {found}, expected {expected}")]
    FetchRemoteMismatch {
        found: RemoteUrl,
        expected: RemoteUrl,
    },
    #[error("IO error accessing repository config")]
    Io(#[from] std::io::Error),
    #[error(transparent)]
    Other(#[from] crate::Error),
}

pub type GitRemoteValidationResult<T> = std::result::Result<T, GitRemoteValidationError>;

#[instrument(skip_all)]
pub fn validate_default_remote(
    open_repository: &dyn OpenRepositoryLike,
    repo_details: &git::RepoDetails,
) -> GitRemoteValidationResult<()> {
    let remote = open_repository
        .find_default_remote()
        .map_err(|_| GitRemoteValidationError::NoDefaultRemote)?;

    let expected_remote_url = repo_details.remote_url().ok_or_else(|| {
        GitRemoteValidationError::Other(err!("Unable to build forge URL from repo details"))
    })?;

    if !expected_remote_url.matches(&remote) {
        return Err(GitRemoteValidationError::RemoteMismatch {
            found: remote,
            expected: expected_remote_url,
        });
    }

    Ok(())
}
