use std::{
    path::{Path, PathBuf},
    sync::{Arc, Mutex},
};

use derive_more::Display;

use secrecy::ExposeSecret as _;

use crate::core::{
    git::{self, Commit, GitRef, RepoDetails},
    BranchName, RemoteUrl,
};
use crate::gitlib::Force;
use crate::prelude::*;

use super::conversions::{as_git2_lock_error, as_git2_open_error};

use git2::{RemoteCallbacks, RemoteUpdateFlags, Repository as Git2Repository};

#[derive(Debug, Display)]
pub struct GitOpenError(crate::Error);
impl std::error::Error for GitOpenError {}

#[derive(Debug, Display, thiserror::Error)]
pub enum GitCloneError {
    Git2(#[from] git2::Error),
    Other(String),
}

impl From<color_eyre::eyre::Error> for GitCloneError {
    fn from(err: color_eyre::eyre::Error) -> Self {
        Self::Other(err.to_string())
    }
}

impl TryFrom<&Path> for GitRepository {
    type Error = Box<GitOpenError>;

    fn try_from(path: &Path) -> Result<Self, Self::Error> {
        Self::open(path)
    }
}

/// An open, thread safe git repository
pub struct GitRepository(Arc<Mutex<Git2Repository>>);

impl std::fmt::Debug for GitRepository {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        let mut debug = f.debug_struct("GitRepository");
        debug.field("path", &self.lock()?.path().to_str());
        debug.finish_non_exhaustive()
    }
}

impl GitRepository {
    pub fn open(path: &Path) -> Result<Self, Box<GitOpenError>> {
        git2::Repository::open(path)
            .map(|repo| Self(Arc::new(Mutex::new(repo))))
            .map_err(as_git2_open_error())
            .map_err(|e| Box::new(GitOpenError(e)))
    }

    pub fn clone(repo_details: &RepoDetails) -> Result<Self, GitCloneError> {
        let repo = git2::Repository::clone(
            repo_details.origin().expose_secret(),
            repo_details.gitdir.pathbuf(),
        )?;
        Ok(Self(Arc::new(Mutex::new(repo))))
    }

    #[instrument]
    pub fn clone_bare(repo_details: &RepoDetails) -> Result<Self, GitCloneError> {
        let mut builder = git2::build::RepoBuilder::new();
        builder.bare(true);
        let repo = builder.clone(
            repo_details.origin().expose_secret(),
            repo_details.gitdir.pathbuf(),
        )?;
        Ok(Self(Arc::new(Mutex::new(repo))))
    }

    pub fn git_dir(&self) -> Result<PathBuf> {
        Ok(self
            .0
            .lock()
            .map(|r| r.path().to_path_buf())
            .map_err(|_| LockError)?)
    }

    pub fn remote_branches(&self) -> Result<Vec<BranchName>> {
        let Ok(repo) = self.0.lock() else {
            return Err(LockError.into());
        };
        repo.config()?.set_bool("core.logAllRefUpdates", true)?;
        repo.odb()?.refresh()?;
        let remote_branches = repo
            .references()?
            .filter_map(Result::ok)
            .filter(git2::Reference::is_remote)
            .filter_map(|r| r.name().map(str::to_string))
            .filter_map(|name| name.strip_prefix("refs/remotes/").map(ToOwned::to_owned))
            .filter(|name| !name.ends_with("HEAD"))
            .map(BranchName::new)
            .collect();
        Ok(remote_branches)
    }

    pub fn find_default_remote_url(&self) -> Result<RemoteUrl> {
        let Ok(repo) = self.0.lock() else {
            return Err(LockError.into());
        };
        let config: git2::Config = repo.config()?;
        let url = config
            .get_string("remote.origin.url")
            .map_err(|_| NoRemoteUrl)?;
        super::GitUrl::from(url).try_into()
    }

    #[allow(clippy::cognitive_complexity)]
    pub fn commit_log(
        &self,
        branch_name: &BranchName,
        limit: usize,
        find_commits: &[git::Commit],
    ) -> Result<Vec<git::Commit>> {
        let Ok(repo) = self.0.lock() else {
            return Err(LockError.into());
        };

        debug!("Getting commit log for branch: {}", branch_name);
        let branch_ref = repo.find_reference(&format!("refs/remotes/origin/{branch_name}"))?;
        let branch_commit = branch_ref.peel_to_commit().map_err(|e| err!(e))?;
        debug!(
            "Branch head commit: id={}, message={:?}",
            branch_commit.id(),
            branch_commit.message().unwrap_or("No message")
        );

        // Examine the repository structure in detail
        debug!("Examining repository structure:");
        let mut examine_walk = repo.revwalk().map_err(|e| err!(e))?;
        examine_walk.push(branch_commit.id()).map_err(|e| err!(e))?;
        examine_walk.set_sorting(git2::Sort::TOPOLOGICAL)?;

        for oid_result in examine_walk {
            let oid = oid_result.map_err(|e| err!(e))?;
            let commit = repo.find_commit(oid).map_err(|e| err!(e))?;

            // Log parent information
            debug!(
                "Commit: id={}, message={:?}, parents={}",
                commit.id(),
                commit.message().unwrap_or("No message"),
                commit.parent_count()
            );

            for i in 0..commit.parent_count() {
                if let Ok(parent) = commit.parent(i) {
                    debug!(
                        "  Parent {}: id={}, message={:?}",
                        i,
                        parent.id(),
                        parent.message().unwrap_or("No message")
                    );
                }
            }
        }

        // Reset and prepare for the actual commit log collection
        let mut revwalk = repo.revwalk().map_err(|e| err!(e))?;
        revwalk.push(branch_commit.id()).map_err(|e| err!(e))?;

        // Configure revwalk to simplify history
        revwalk.simplify_first_parent().map_err(|e| err!(e))?;

        debug!("Walking commit history (limit={})", limit);
        let mut commits = vec![];
        let mut seen_shas = std::collections::HashSet::new();

        for (i, oid_result) in revwalk.enumerate().take(limit) {
            let oid = oid_result.map_err(|e| err!(e))?;

            // Skip if we've already seen this commit
            if !seen_shas.insert(oid.to_string()) {
                debug!("Skipping duplicate commit: {}", oid);
                continue;
            }

            let commit = repo.find_commit(oid).map_err(|e| err!(e))?;
            let sha = git::commit::Sha::new(commit.id().to_string());
            let message =
                git::commit::Message::new(commit.message().unwrap_or_default().to_string());

            debug!(
                "Commit [{}]: id={}, message={:?}",
                i,
                oid,
                commit.message().unwrap_or("No message")
            );

            let git_commit = git::Commit::new(sha, message);
            if find_commits.contains(&git_commit) {
                debug!("Found target commit: {}", oid);
                commits.push(git_commit);
                break;
            }
            commits.push(git_commit);
        }

        debug!("Found {} commits in history", commits.len());
        Ok(commits)
    }

    pub fn read_file(&self, branch_name: &BranchName, file_name: &Path) -> Result<String> {
        let Ok(repo) = self.0.lock() else {
            return Err(LockError.into());
        };
        let reference = repo
            .find_reference(&format!("refs/remotes/origin/{branch_name}"))
            .map_err(|e| err!("Failed to find remote branch: {}", e.message()))?;
        let commit = reference
            .peel_to_commit()
            .map_err(|e| err!("Failed to resolve commit: {}", e.message()))?;
        let tree = commit
            .tree()
            .map_err(|e| err!("Failed to get commit tree: {}", e.message()))?;
        let entry = tree
            .get_path(file_name)
            .map_err(|_| git::file::Error::FileNotFound)?;
        let object = entry
            .to_object(&repo)
            .map_err(|e| err!("Failed to get file object: {}", e.message()))?;
        let blob = object.as_blob().ok_or(git::file::Error::FileNotFound)?;
        std::str::from_utf8(blob.content())
            .map(std::string::ToString::to_string)
            .map_err(|e| err!("UTF-8 conversion failed: {}", e))
    }

    pub fn fetch(&self) -> Result<()> {
        let Ok(repo) = self.0.lock() else {
            return Err(LockError.into());
        };
        let mut remote = repo
            .find_remote("origin")
            .map_err(|e| err!("git2: failed to find remote 'origin': {}", e.message()))?;
        let mut fetch_options = git2::FetchOptions::new();
        fetch_options.prune(git2::FetchPrune::On);
        let specs: [&str; 0] = [];
        remote
            .download(&specs, Some(&mut fetch_options))
            .map_err(|e| err!("git2: fetch failed: {}", e.message()))?;
        remote
            .update_tips(
                None,
                RemoteUpdateFlags::UPDATE_FETCHHEAD,
                git2::AutotagOption::None,
                None,
            )
            .map_err(|e| err!("git2: failed to update refs: {}", e.message()))?;
        Ok(())
    }

    #[instrument]
    pub fn push(&self, commit_id: &GitRef, target_branch: &BranchName, force: Force) -> Result<()> {
        let Ok(repo) = self.0.lock() else {
            return Err(LockError.into());
        };
        let mut push_options = git2::PushOptions::new();

        // Create the refspec with or without force flag
        let refspec = match &force {
            Force::None => format!("{commit_id}:refs/heads/{target_branch}"),
            Force::ForceWithLease { .. } => format!("+{commit_id}:refs/heads/{target_branch}"),
        };

        debug!("Refspec for push: {}", refspec);

        // Set up callbacks for force-with-lease
        if let Force::ForceWithLease { ref expected_sha } = force {
            let mut callbacks = RemoteCallbacks::new();
            let expected_sha = expected_sha.clone(); // Clone to move into the closure
            callbacks.push_negotiation(move |updates| {
                for update in updates {
                    debug!(
                        "Push negotiation: dst_refname={:?}, dst={:?}, src={:?}",
                        update.dst_refname(),
                        update.dst(),
                        update.src()
                    );

                    if update.dst_refname() == Some(&format!("refs/remotes/origin/{target_branch}"))
                        && update.dst().to_string() != expected_sha
                    {
                        error!(
                            "Remote branch has diverged! Expected: {}, Actual: {}",
                            expected_sha,
                            update.dst().to_string()
                        );
                        return Err(git2::Error::from_str("Remote branch has diverged"))?;
                    }
                }
                Ok(())
            });
            push_options.remote_callbacks(callbacks);
        };

        let mut remote = repo
            .find_remote("origin")
            .map_err(|e| err!("git2: failed to find remote 'origin': {}", e.message()))?;

        debug!(
            "git2 push of '{commit_id}' to branch '{target_branch}' starting with force={:?}",
            force
        );

        match remote.push(&[&refspec], Some(&mut push_options)) {
            Ok(()) => {
                debug!(
                    "git2 push of '{commit_id}' to branch '{target_branch}' completed successfully"
                );
                Ok(())
            }
            Err(e) => {
                error!("git2 push failed: {}", e);

                // Try to get more diagnostic information
                if let Ok(config) = repo.config() {
                    if let Ok(url) = config.get_string("remote.origin.url") {
                        debug!("Remote URL: {}", url);
                    }
                }

                // Check if the branch exists
                if let Ok(reference) =
                    repo.find_reference(&format!("refs/remotes/origin/{target_branch}"))
                {
                    if let Ok(target_commit) = reference.peel_to_commit() {
                        debug!("Current remote branch commit: {}", target_commit.id());
                    }
                } else {
                    debug!("Remote branch not found locally");
                }

                Err(err!("git2: push failed: {}", e.message()))
            }
        }
    }

    pub fn switch_branch(&self, branch_name: &str) -> Result<()> {
        debug!("Switching to branch '{branch_name}' via gitlib");
        let repo = self.0.lock().map_err(as_git2_lock_error())?;
        let branch_name_ref = format!("refs/heads/{branch_name}");
        let _ = repo.revparse_ext(branch_name)?; // Verify the branch exists
        repo.set_head(&branch_name_ref)?;
        repo.checkout_head(Some(git2::build::CheckoutBuilder::new().force()))?;
        drop(repo);
        debug!("Switch branch to '{branch_name}' completed");
        Ok(())
    }

    #[allow(clippy::too_many_lines)]
    pub fn rebase_onto(&self, new_base: &str, old_base: &str, branch: &str) -> Result<()> {
        debug!("Rebasing branch '{branch}' onto '{new_base}' (old base: '{old_base}') via gitlib");
        #[allow(clippy::cognitive_complexity)]
        self.0.lock().map_err(|_| err!(LockError)).and_then(|repo| {
            let commit_new = repo
                .revparse_single(new_base)
                .map_err(|e| err!("Invalid new base reference '{new_base}': {}", e))?;
            let annotated_new = repo.find_annotated_commit(commit_new.id())?;
            debug!("New base commit: id={}, message={:?}", 
                commit_new.id(),
                commit_new.as_commit().map_or("Not a commit", |c| c.message().unwrap_or("No message")));
            let commit_old = repo
                .revparse_single(old_base)
                .map_err(|e| err!("Invalid old base reference '{old_base}': {}", e))?;
            let annotated_old = repo.find_annotated_commit(commit_old.id())?;
            debug!("Old base commit: id={}, message={:?}", 
                commit_old.id(),
                commit_old.as_commit().map_or("Not a commit", |c| c.message().unwrap_or("No message")));
            let commit_branch = repo
                .revparse_single(branch)
                .map_err(|e| err!("Invalid branch reference '{branch}': {}", e))?;
            let annotated_branch = repo.find_annotated_commit(commit_branch.id())?;
            debug!("Branch head commit: id={}, message={:?}", 
                commit_branch.id(),
                commit_branch.as_commit().map_or("Not a commit", |c| c.message().unwrap_or("No message")));

            // Skip rebase if old and new base are the same
            if commit_new.id() == commit_old.id() {
                debug!("New base and old base are the same, no rebase needed");
                return Ok(());
            }

            // Log the parent relationships
            if let Some(branch_commit) = commit_branch.as_commit() {
                debug!("Branch commit has {} parents", branch_commit.parent_count());
                for i in 0..branch_commit.parent_count() {
                    if let Ok(parent) = branch_commit.parent(i) {
                        debug!("Branch parent {}: id={}, message={:?}", 
                            i, parent.id(), parent.message().unwrap_or("No message"));
                    }
                }
            }

            // Create a new rebase operation
            let mut rebase_options = git2::RebaseOptions::new();
            debug!("Starting rebase operation: branch={} onto={} from={}", 
                branch, commit_new.id(), commit_old.id());
            let mut rebase = repo
                .rebase(
                    Some(&annotated_branch),
                    Some(&annotated_old),
                    Some(&annotated_new),
                    Some(&mut rebase_options),
                )
                .map_err(|e| err!("Rebase initialization failed: {}", e))?;

            let sig = repo.signature()?;
            let mut last_rebased_commit_id: Option<git2::Oid> = None;
            let mut operation_count = 0;

            // Process each commit in the rebase
            while let Some(operation_result) = rebase.next() {
                operation_count += 1;
                let operation = operation_result.map_err(|e| {
                    let _ = rebase.abort();
                    err!("Error getting next rebase operation: {}", e)
                })?;
                debug!("Rebase operation {}: commit_id={}", operation_count, operation.id());
                if let Ok(commit) = repo.find_commit(operation.id()) {
                    debug!("Rebasing commit: id={}, message={:?}", 
                        commit.id(), commit.message().unwrap_or("No message"));
                }

                match rebase.commit(None, &sig, None) {
                    Ok(new_commit_id) => {
                        if let Ok(new_commit) = repo.find_commit(new_commit_id) {
                            debug!("Successfully rebased commit: old_id={}, new_id={}, message={:?}", 
                                operation.id(), new_commit_id, new_commit.message().unwrap_or("No message"));
                            // Log parent information for the new commit
                            debug!("New commit has {} parents", new_commit.parent_count());
                            for i in 0..new_commit.parent_count() {
                                if let Ok(parent) = new_commit.parent(i) {
                                    debug!("New commit parent {}: id={}, message={:?}", 
                                        i, parent.id(), parent.message().unwrap_or("No message"));
                                }
                            }
                        } else {
                            debug!("Successfully rebased commit: old_id={}, new_id={}", 
                                operation.id(), new_commit_id);
                        }
                        last_rebased_commit_id = Some(new_commit_id);
                    }
                    Err(e) => {
                        if e.code() == git2::ErrorCode::Applied && e.class() == git2::ErrorClass::Rebase {
                            debug!("Commit already applied: id={}", operation.id());
                            // This commit was already applied, continue to next
                            continue;
                        }
                        error!("Failed to apply commit {} during rebase: {}", operation.id(), e);
                        if let Err(abort_err) = rebase.abort() {
                            error!("Failed to abort rebase after commit failure: {}", abort_err);
                            return Err(err!(
                                "Rebase conflict or error applying commit {}. Rebase aborted uncleanly. Apply Error: {}. Abort Error: {}",
                                operation.id(), e, abort_err
                            ));
                        }
                        return Err(err!(
                            "Rebase conflict or error applying commit {}. Rebase aborted. Error: {}",
                            operation.id(), e
                        ));
                    }
                }
            }

            debug!("Rebase operations completed: {} operations processed", operation_count);

            // Use the last rebased commit if we have one, otherwise use the new base
            let final_target_oid = last_rebased_commit_id.unwrap_or_else(|| commit_new.id());
            debug!("Final target commit: id={}", final_target_oid);

            // Finish the rebase operation
            rebase.finish(Some(&sig))
                .map_err(|e| err!("Rebase cleanup failed: {}", e))?;
            debug!("Rebase finished successfully");

            let branch_ref_name = format!("refs/heads/{branch}");
            let remote_ref_name = format!("refs/remotes/origin/{branch}");

            // Update the local branch reference
            repo.reference(
                &branch_ref_name,
                final_target_oid,
                true,
                &format!("rebase: {branch} onto {new_base}"),
            )?;
            debug!("Updated local branch reference: {}", branch_ref_name);

            // Update the remote tracking reference
            repo.reference(
                &remote_ref_name,
                final_target_oid,
                true,
                "Update remote ref after rebase",
            )?;
            debug!("Updated remote tracking reference: {}", remote_ref_name);

            // Log the final commit history
            debug!("Final commit history after rebase:");
            let mut revwalk = repo.revwalk()?;
            revwalk.push(final_target_oid)?;
            for (i, oid_result) in revwalk.enumerate().take(10) {
                if let Ok(oid) = oid_result {
                    if let Ok(commit) = repo.find_commit(oid) {
                        debug!("  [{}] {}: {}", i, oid, commit.message().unwrap_or("No message"));
                    } else {
                        debug!("  [{}] {}: <not a commit>", i, oid);
                    }
                }
            }

            Ok(())
        })
    }

    #[cfg(test)]
    pub fn init(path: &Path) -> Result<Self, Box<git_init::GitInitError>> {
        git2::Repository::init_opts(
            path,
            git2::RepositoryInitOptions::new().bare(false).mkpath(true),
        )
        .map_err(|e| Box::new(git_init::GitInitError::Init(e)))?;
        Self::open(path).map_err(|e| Box::new(git_init::GitInitError::Open(*e)))
    }

    pub fn lock(&self) -> Result<std::sync::MutexGuard<'_, git2::Repository>, LockError> {
        self.0.lock().map_err(|_| LockError)
    }

    pub fn new(repo: Git2Repository) -> Self {
        Self(Arc::new(Mutex::new(repo)))
    }
}

#[derive(Debug, Display, thiserror::Error)]
pub struct NoRemoteFound;

#[derive(Debug, Display, thiserror::Error)]
pub struct NoRemoteUrl;

#[derive(Debug, Display, thiserror::Error)]
pub struct LockError;

impl From<LockError> for std::fmt::Error {
    fn from(_value: LockError) -> Self {
        Self
    }
}

#[cfg(test)]
mod git_init {
    #[derive(Debug, derive_more::Display)]
    pub enum GitInitError {
        Init(git2::Error),
        Open(super::GitOpenError),
    }
    impl std::error::Error for GitInitError {}
}

fn add_file_and_commit(
    repo: &GitRepository,
    branch_name: &str,
    file_path: &Path,
    content: &str,
) -> Result<Commit> {
    repo.0.lock().map_err(|_| err!(LockError)).and_then(|repo| {
        let branch_ref_name = format!("refs/heads/{branch_name}");
        let parent_ref = repo.find_reference(&branch_ref_name)?;
        let parent_commit = parent_ref.peel_to_commit()?;
        let mut index = repo.index()?;
        #[allow(clippy::expect_used)]
        let workdir = repo.workdir().expect("git workdir");
        let absolute_path = workdir.join(file_path);
        if let Some(parent_dir) = absolute_path.parent() {
            std::fs::create_dir_all(parent_dir)?;
        }
        std::fs::write(&absolute_path, content)?;
        index.add_path(file_path)?;
        let tree_id = index.write_tree()?;
        let tree = repo.find_tree(tree_id)?;
        let sig = git2::Signature::now("Test", "test@example.com")?;
        let commit_message = format!("Commit {content}");
        let commit_id = repo.commit(
            Some(&branch_ref_name),
            &sig,
            &sig,
            &commit_message,
            &tree,
            &[&parent_commit],
        )?;
        let remote_ref_name = format!("refs/remotes/origin/{branch_name}");
        let _ = repo.reference(&remote_ref_name, commit_id, true, &commit_message);
        repo.checkout_head(Some(git2::build::CheckoutBuilder::new().force()))?;
        debug!(%branch_name, ?file_path, %commit_id, %commit_message, "committed file");
        Ok(Commit::new(
            git::commit::Sha::new(commit_id.to_string()),
            git::commit::Message::new(commit_message),
        ))
    })
}

#[cfg(test)]
mod tests {
    use git::Commit;

    use crate::gitlib::tests::RemoteTestSetup;

    use super::*;

    #[allow(clippy::cognitive_complexity)]
    fn create_test_repo_with_remote() -> Result<(RemoteTestSetup, GitRepository, Commit)> {
        let setup = RemoteTestSetup::new()?;
        let repo = setup.clone_local_repo()?;

        // Log the repository state before making any changes
        debug!("Repository state before setup:");
        let mut initial_commit_id = None;
        let mut initial_commit_message = None;

        if let Ok(head) = repo.head() {
            if let Ok(head_commit) = head.peel_to_commit() {
                debug!(
                    "HEAD is at commit: id={}, message={:?}",
                    head_commit.id(),
                    head_commit.message().unwrap_or("No message")
                );

                // Store the initial commit details to reuse
                initial_commit_id = Some(head_commit.id());
                initial_commit_message = head_commit.message().map(String::from);

                // Check the initial repository history
                let mut revwalk = repo.revwalk()?;
                revwalk.push(head_commit.id())?;
                debug!("Initial repository history:");
                for (i, oid_result) in revwalk.enumerate().take(5) {
                    if let Ok(oid) = oid_result {
                        if let Ok(commit) = repo.find_commit(oid) {
                            debug!(
                                "  [{}] {}: {}",
                                i,
                                oid,
                                commit.message().unwrap_or("No message")
                            );
                        }
                    }
                }
            }
        }

        // Use the existing initial commit instead of creating a new one
        let (commit_id, commit_message) =
            if let (Some(id), Some(msg)) = (initial_commit_id, initial_commit_message) {
                debug!("Using existing initial commit: id={}, message={}", id, msg);
                (id, msg)
            } else {
                // Fallback to creating a new commit if needed
                let sig = repo.signature()?;
                let msg = s!("Initial commit");
                let id = repo.commit(
                    Some("refs/heads/main"),
                    &sig,
                    &sig,
                    &msg,
                    &repo.find_tree(repo.treebuilder(None)?.write()?)?,
                    &[], // No parents since we're creating a new initial commit
                )?;
                debug!("Created new initial commit: id={}, message={}", id, msg);
                (id, msg)
            };

        // Create dev branch pointing to the initial commit
        repo.branch("dev", &repo.find_commit(commit_id)?, false)?;
        debug!("Created dev branch pointing to initial commit");

        // Set HEAD to main
        repo.set_head("refs/heads/main")?;
        debug!("Set HEAD to main branch");

        {
            // Explicitly set config for branch remote and merge
            let mut config = repo.config()?;
            config.set_str("branch.main.remote", "origin")?;
            config.set_str("branch.main.merge", "refs/heads/main")?;
            config.set_str("branch.main.rebase", "true")?;
            // Also configure dev branch
            config.set_str("branch.dev.remote", "origin")?;
            config.set_str("branch.dev.merge", "refs/heads/dev")?;
            config.set_str("branch.dev.rebase", "true")?;
            debug!("Configured branch tracking information");
        }

        // Log the repository state after setup
        debug!("Repository state after setup:");
        let mut revwalk = repo.revwalk()?;
        revwalk.push_head()?;
        for (i, oid_result) in revwalk.enumerate().take(5) {
            if let Ok(oid) = oid_result {
                if let Ok(commit) = repo.find_commit(oid) {
                    debug!(
                        "  [{}] {}: {}",
                        i,
                        oid,
                        commit.message().unwrap_or("No message")
                    );
                }
            }
        }

        // Re-open the repository to ensure config changes are loaded consistently
        Ok((setup, GitRepository::open(repo.path())?, {
            git::Commit::builder()
                .sha(git::commit::Sha::new(commit_id.to_string()))
                .message(git::commit::Message::new(commit_message))
                .build()
        }))
    }

    mod git_init {

        use super::*;

        #[test]
        fn test_open_git2() -> Result<()> {
            let temp_dir = tempfile::tempdir()?;
            let repo_path = temp_dir.path();
            git2::Repository::init(repo_path)?;
            let repo = GitRepository::open(repo_path);
            assert!(repo.is_ok());
            Ok(())
        }
    }

    mod clone_tests {
        use assert2::let_assert;

        use crate::repo::tests::given;

        use super::*;

        // Compile-time interface verification
        #[allow(dead_code)]
        fn test_clone_interface() {
            let _ = |repo_details: &RepoDetails| -> Result<GitRepository, GitCloneError> {
                GitRepository::clone(repo_details)
            };
        }

        // Error type propagation check
        #[test]
        fn errors_propagate_correctly() {
            if false {
                let fs = given::a_filesystem();
                let dummy_details = given::repo_details(&fs);
                let_assert!(Err(git2_err) = GitRepository::clone(&dummy_details));
                assert!(matches!(git2_err, GitCloneError::Git2(_)));
            }
        }
    }

    mod remote_branches_tests {
        use super::*;

        use tempfile::tempdir;

        fn create_test_repo(dir: &tempfile::TempDir) -> Result<GitRepository> {
            let repo = GitRepository::init(dir.path())?;
            repo.0
                .lock()
                .map_err(|_| err!("repo lock"))
                .and_then(|repo| {
                    let refs_dir = repo.path().join("refs/remotes/origin");
                    std::fs::create_dir_all(&refs_dir)?;
                    let objects_dir = repo.path().join("objects");
                    std::fs::create_dir_all(&objects_dir)?;
                    let sig = repo.signature()?;
                    let tree_id = repo.treebuilder(None)?.write()?;
                    let commit_id_main = repo.commit(
                        Some("refs/heads/main"),
                        &sig,
                        &sig,
                        "Initial commit",
                        &repo.find_tree(tree_id)?,
                        &[],
                    )?;
                    repo.reference("refs/remotes/origin/main", commit_id_main, true, "test")?;
                    let commit_id_dev = repo.commit(
                        Some("refs/heads/dev"),
                        &sig,
                        &sig,
                        "Dev commit",
                        &repo.find_tree(repo.treebuilder(None)?.write()?)?,
                        &[],
                    )?;
                    repo.reference("refs/remotes/origin/dev", commit_id_dev, true, "test")?;
                    repo.odb()?.refresh()?;
                    Ok(())
                })?;

            Ok(repo)
        }

        #[test_log::test]
        fn commit_log_returns_expected_commits() -> Result<()> {
            let (_setup, repo, test_commit) = create_test_repo_with_remote()?;
            debug!("[1]");
            repo.push(
                &GitRef::from(test_commit.clone()),
                &BranchName::from(s!("main")),
                Force::None,
            )?;
            debug!("[2]");

            let commits = repo.commit_log(&BranchName::new("main"), 10, &[])?;
            assert!(!commits.is_empty(), "No commits found");

            let limited = repo.commit_log(&BranchName::new("main"), 1, &[])?;
            assert_eq!(limited.len(), 1, "Limit not respected");

            let found = repo.commit_log(&BranchName::new("main"), 10, &[test_commit.clone()])?;
            assert!(found.contains(&test_commit), "Commit not found");
            Ok(())
        }

        #[test]
        fn lists_remote_branches_correctly() -> Result<()> {
            let dir = tempdir()?;
            let repo = create_test_repo(&dir)?;

            let mut branches = repo.remote_branches()?;
            branches.sort_by_key(|b| b.as_str().to_string());

            assert_eq!(
                branches.iter().map(|b| b.as_str()).collect::<Vec<_>>(),
                vec!["origin/dev", "origin/main"],
                "Failed",
            );
            Ok(())
        }

        #[test]
        fn filters_out_origin_head() -> Result<()> {
            let dir = tempdir()?;
            let repo = create_test_repo(&dir)?;

            // Add HEAD reference
            repo.0
                .lock()
                .map_err(|_| err!("repo lock"))
                .and_then(|repo| {
                    // Create valid commit for HEAD
                    let sig = repo.signature()?;
                    let commit_id = repo.commit(
                        Some("refs/heads/HEAD"),
                        &sig,
                        &sig,
                        "HEAD commit",
                        &repo.find_tree(repo.treebuilder(None)?.write()?)?,
                        &[],
                    )?;

                    repo.reference("refs/remotes/origin/HEAD", commit_id, true, "test")?;
                    Ok(())
                })?;

            let branches = repo.remote_branches()?;
            assert!(!branches.iter().any(|b| b.as_str() == "HEAD"));
            Ok(())
        }

        #[test]
        fn handles_special_characters() -> Result<()> {
            let special_branch = "feat/123-special!@#";

            let dir = tempdir()?;
            let repo = create_test_repo(&dir)?;

            repo.0
                .lock()
                .map_err(|_| err!("repo lock"))
                .and_then(|repo| {
                    // Create valid commit for special branch
                    let sig = repo.signature()?;
                    let commit_id = repo.commit(
                        Some(&format!("refs/heads/{special_branch}")),
                        &sig,
                        &sig,
                        "Special commit",
                        &repo.find_tree(repo.treebuilder(None)?.write()?)?,
                        &[],
                    )?;

                    repo.reference(
                        &format!("refs/remotes/origin/{special_branch}"),
                        commit_id,
                        true,
                        "test",
                    )?;
                    Ok(())
                })?;

            let branches = repo.remote_branches()?;
            assert!(
                branches
                    .iter()
                    .any(|b| b.as_str().ends_with(special_branch)),
                "Special branch not found"
            );
            Ok(())
        }
    }

    mod find_default_remote_tests {
        use super::*;

        #[test]
        fn finds_origin_remote_url() -> Result<()> {
            let (setup, repo, _) = create_test_repo_with_remote()?;

            let url = repo.find_default_remote_url()?;
            assert_eq!(
                url.original_url.strip_suffix("/"),
                setup
                    .remote_repo()?
                    .path()
                    .to_string_lossy()
                    .strip_prefix("file://"),
                "Failed"
            );
            Ok(())
        }

        #[test]
        fn handles_missing_remote_url() -> Result<()> {
            let (_setup, repo, _) = create_test_repo_with_remote()?;

            // Modify the config to remove the URL for the 'origin' remote
            repo.lock()?.config()?.remove("remote.origin.url")?;
            let result = repo.find_default_remote_url();
            assert!(
                matches!(result.err(), Some(e) if e.downcast_ref::<NoRemoteUrl>().is_some()),
                "Git2 backend failed: Expected NoRemoteUrl error"
            );
            Ok(())
        }
    }

    mod read_file_tests {
        use super::*;
        use assert2::let_assert;

        mod rebase_tests {
            use super::*;

            fn create_branch(repo: &GitRepository, branch_name: &str) -> Result<()> {
                let repo = repo.lock()?;
                {
                    let head = repo.head()?.peel_to_commit()?;
                    repo.branch(branch_name, &head, false)?;
                }
                drop(repo);
                Ok(())
            }

            #[test_log::test]
            fn basic_rebase_onto_new_base() -> Result<()> {
                let (_setup, repo, initial_commit_obj) = create_test_repo_with_remote()?;
                let initial_commit_sha = initial_commit_obj.sha().as_str();

                // Create feature branch from initial commit
                create_branch(&repo, "feature")?;
                // Add commit to feature branch
                let _feature_commit = add_file_and_commit(
                    &repo,
                    "feature",
                    Path::new("feature-file.txt"),
                    "feature1",
                )?;
                // Add commit to main branch
                let main1_commit =
                    add_file_and_commit(&repo, "main", Path::new("main-file.txt"), "main1")?;

                debug!(
                    "Rebasing feature branch onto main ({}) using old base ({})",
                    main1_commit.sha(),
                    initial_commit_sha
                );
                // Rebase feature onto the new main commit, specifying the initial commit as the old base
                repo.rebase_onto(main1_commit.sha().as_str(), initial_commit_sha, "feature")?;

                let log = repo.commit_log(&BranchName::new("feature"), 10, &[])?;
                let messages: Vec<_> = log.iter().map(|c| c.message().as_str()).collect();
                assert_eq!(
                    messages,
                    ["Commit feature1", "Commit main1", "Initial commit"],
                    "Rebase failed to reorder commits correctly"
                );
                Ok(())
            }

            #[test]
            fn detects_rebase_conflicts() -> Result<()> {
                // Use _initial_commit as we don't need it directly here
                let (_setup, repo, _initial_commit) = create_test_repo_with_remote()?;

                // Base commit
                let base_commit =
                    add_file_and_commit(&repo, "main", Path::new("file.txt"), "base")?;

                // Create conflicting changes
                create_branch(&repo, "feature")?;
                // Commit on feature branch
                let _feature_commit =
                    add_file_and_commit(&repo, "feature", Path::new("file.txt"), "feature1")?;
                // Conflicting commit on main branch
                let _conflicting_commit =
                    add_file_and_commit(&repo, "main", Path::new("file.txt"), "conflicting")?;

                // Attempt rebase: feature onto main, using base_commit as the point after which feature diverged
                let result = repo.rebase_onto("main", base_commit.sha().as_str(), "feature");
                assert!(result.is_err(), "Should error on conflict during rebase");

                if let Err(e) = result {
                    assert!(
                        e.to_string().contains("conflict")
                            || e.to_string().contains("Failed to apply"), // Git2 error message might vary slightly
                        "Unexpected error type: {e}"
                    );
                }

                Ok(())
            }

            #[test]
            fn handles_non_linear_history() -> Result<()> {
                let (_setup, repo, _initial_commit) = create_test_repo_with_remote()?;
                // starts with main and dev branches on the same commit

                // Create complex history
                // Commit on main
                let main1_commit =
                    add_file_and_commit(&repo, "main", Path::new("file1.txt"), "main1")?;
                // Commit on dev
                let _dev1_commit =
                    add_file_and_commit(&repo, "dev", Path::new("file2.txt"), "dev1")?;
                // Another commit on main
                let _main2_commit =
                    add_file_and_commit(&repo, "main", Path::new("file3.txt"), "main2")?;

                // Rebase dev onto main, specifying main1 as the old base (the point dev branched off)
                repo.rebase_onto("main", main1_commit.sha().as_str(), "dev")?;

                // --- Simulate push: Update remote tracking branch manually ---
                let repo_lock = repo.lock()?;
                {
                    // Find the new HEAD of the local 'dev' branch after rebase
                    let local_dev_ref = repo_lock.find_reference("refs/heads/dev")?;
                    let new_dev_head_oid = local_dev_ref
                        .target()
                        .ok_or_else(|| err!("Local dev ref has no target OID"))?;
                    // Update the remote tracking ref to point to the new local HEAD
                    repo_lock.reference(
                        "refs/remotes/origin/dev",
                        new_dev_head_oid,
                        true, // Force update
                        "test: simulate push after rebase",
                    )?;
                }
                drop(repo_lock);

                let log = repo.commit_log(&BranchName::new("dev"), 10, &[])?;
                let messages: Vec<_> = log.iter().map(|c| c.message().as_str()).collect();
                assert_eq!(
                    messages,
                    [
                        "Commit dev1",
                        "Commit main2",
                        "Commit main1",
                        "Initial commit"
                    ],
                    "Rebase didn't maintain commit order correctly"
                );

                Ok(())
            }

            #[test]
            fn errors_on_invalid_refs() -> Result<()> {
                let (_setup, repo, _initial_commit) = create_test_repo_with_remote()?;
                let base_commit =
                    add_file_and_commit(&repo, "main", Path::new("file.txt"), "base")?;
                create_branch(&repo, "feature")?;

                // Test invalid new base
                let result = repo.rebase_onto("invalid_ref", base_commit.sha().as_str(), "feature");
                assert!(result.is_err(), "Rebase should fail with invalid new base");

                // Test invalid old base
                let result = repo.rebase_onto("main", "invalid_ref", "feature");
                assert!(result.is_err(), "Rebase should fail with invalid old base");

                // Test invalid branch
                let result = repo.rebase_onto("main", base_commit.sha().as_str(), "invalid_branch");
                assert!(result.is_err(), "Rebase should fail with invalid branch");

                Ok(())
            }

            #[test]
            fn maintains_commit_metadata() -> Result<()> {
                let (_setup, repo, _initial_commit) = create_test_repo_with_remote()?;

                // Base commit on main
                let base_commit =
                    add_file_and_commit(&repo, "main", Path::new("file.txt"), "base")?;

                // Create feature branch
                create_branch(&repo, "feature")?;

                // Commit on feature branch
                let original_commit =
                    add_file_and_commit(&repo, "feature", Path::new("file.txt"), "feature1")?;

                // Add another commit to main to advance it
                let _main_advance_commit = add_file_and_commit(
                    &repo,
                    "main",
                    Path::new("another-file.txt"),
                    "main advance",
                )?;

                // Now rebase feature onto the new main
                repo.rebase_onto("main", base_commit.sha().as_str(), "feature")?;

                // --- Simulate push: Update remote tracking branch manually ---
                let repo_lock = repo.lock()?;
                {
                    // Find the new HEAD of the local 'feature' branch after rebase
                    let local_feature_ref = repo_lock.find_reference("refs/heads/feature")?;
                    let new_feature_head_oid = local_feature_ref
                        .target()
                        .ok_or_else(|| err!("Local feature ref has no target OID"))?;
                    // Update the remote tracking ref to point to the new local HEAD
                    repo_lock.reference(
                        "refs/remotes/origin/feature",
                        new_feature_head_oid,
                        true, // Force update
                        "test: simulate push after rebase",
                    )?;
                }
                drop(repo_lock);

                // Get the log after rebase. We expect 2 commits on feature now relative to base:
                // the rebased 'feature1' and the 'base' commit itself.
                let rebased_commits = repo.commit_log(&BranchName::new("feature"), 2, &[])?;
                assert_eq!(
                    rebased_commits.len(),
                    2,
                    "Expected 2 commits after rebase onto base"
                );

                // The first commit in the log should be the rebased one
                let rebased_feature_commit = &rebased_commits[0];

                assert_eq!(
                    rebased_feature_commit.message(),
                    original_commit.message(),
                    "Commit message changed during rebase"
                );
                // SHA *must* change after rebase
                assert_ne!(
                    rebased_feature_commit.sha(),
                    original_commit.sha(),
                    "Commit SHA should change during rebase"
                );

                Ok(())
            }
        }

        #[test]
        fn reads_existing_file_from_branch() -> Result<()> {
            let (_setup, repo, _initial_commit) = create_test_repo_with_remote()?;

            // Create test file
            let file_path = Path::new("test.txt");
            let expected_content = "test content";

            let _commit = add_file_and_commit(&repo, "main", file_path, expected_content)?;

            let content = repo.read_file(&BranchName::new("main"), file_path)?;
            assert_eq!(content, expected_content);
            Ok(())
        }

        #[test]
        fn returns_error_for_missing_file() -> Result<()> {
            let (_setup, repo, _initial_commit) = create_test_repo_with_remote()?;

            let file_path = Path::new("test.txt");
            let expected_content = "test content";

            // Use the refactored helper
            let _commit = add_file_and_commit(&repo, "main", file_path, expected_content)?;

            let result = repo.read_file(&BranchName::new("main"), Path::new("missing.txt"));
            let_assert!(Err(e) = result);
            assert!(matches!(
                e.downcast_ref::<git::file::Error>(),
                Some(git::file::Error::FileNotFound)
            ));
            Ok(())
        }

        #[test]
        fn handles_special_characters_in_filename() -> Result<()> {
            let (_setup, repo, _initial_commit) = create_test_repo_with_remote()?;

            let file_path = Path::new("spécial file@123.txt");
            let expected_content = "special content";

            // Use the refactored helper
            let _commit = add_file_and_commit(&repo, "main", file_path, expected_content)?;

            let content = repo.read_file(&BranchName::new("main"), file_path)?;
            assert_eq!(content, expected_content);
            Ok(())
        }
    }
}
