//
#![allow(dead_code)]
#![allow(clippy::module_name_repetitions)]

use kameo::actor::ActorRef;
use kameo_actors::pubsub::PubSub;
use server::server_mode::ServerMode;

use std::path::PathBuf;

use clap::Parser;

pub use io_send_error::IoSendError;
pub use io_send_error::MessageResult;

use kxio::{fs, net};

use crate::core::git;
use crate::prelude::*;

mod alerts;
mod base_actor;
mod core;
mod file_watcher;
mod forge;
mod forges;
mod gitlib;
mod init;
mod io_send_error;
mod macros;
mod prelude;
mod repo;
mod server;
mod tui;
mod webhook;
mod xui;

#[cfg(test)]
mod tests;

pub type MessageBus<T> = ActorRef<PubSub<T>>;

#[derive(Parser, Debug)]
#[clap(version = clap::crate_version!(), author = clap::crate_authors!(), about = clap::crate_description!())]
struct Commands {
    #[clap(subcommand)]
    command: Command,
}
#[derive(Parser, Debug)]
enum Command {
    Init,
    #[clap(subcommand)]
    Server(Server),
}
#[derive(Parser, Debug)]
enum Server {
    Init,
    Start {
        /// Display a UI
        #[arg(long, required = false, group = "mode")]
        ui: bool,

        /// Display a UI (experimental)
        #[arg(long, required = false, group = "mode")]
        xui: bool,
    },
}

fn main() -> Result<()> {
    color_eyre::install()?;

    let fs = fs::new(PathBuf::default());
    let net = net::new();
    let repository_factory = git::repository::factory::real();
    let commands = Commands::parse();

    match commands.command {
        Command::Init => init::run(&fs),
        Command::Server(server) => match server {
            Server::Init => server::init(&fs),
            Server::Start { ui, xui } => {
                let server_mode = match (ui, xui) {
                    (true, true) => unreachable!(), // should be disallowed by 'group = "mode"'
                    (true, false) => ServerMode::LegacyUI,
                    (false, true) => ServerMode::NewUI,
                    (false, false) => ServerMode::Simple,
                };
                server::start(
                    server_mode,
                    &fs,
                    &net,
                    repository_factory,
                    std::time::Duration::from_secs(10),
                )
            }
        },
    }
}
