use crate::prelude::*;
use crate::{
    core::{
        git::{
            commit::Message, push::reset, repository::open::OpenRepositoryLike, Commit, RepoDetails,
        },
        RepoConfig,
    },
    gitlib::Force,
    repo::messages::MessageToken,
};

use serde::{Deserialize, Serialize};

#[derive(Clone, Debug, PartialEq, Eq, derive_more::Display, Serialize, Deserialize)]
pub enum AdvanceError {
    /// The next commit message is invalid (non-conventional)
    UnconventionalCommitMessage { reason: String },
    /// The next commit has a 'WIP:' commit message prefix
    NextCommitIsWIP,
    /// Unexpected error
    Unexpected(String),
}

// advance next to the next commit towards the head of the dev branch
#[instrument(fields(next), skip_all)]
pub fn advance_next(
    commit: Commit,
    force: Force,
    repo_details: &RepoDetails,
    repo_config: &RepoConfig,
    open_repository: &dyn OpenRepositoryLike,
    message_token: MessageToken,
) -> Result<MessageToken, AdvanceError> {
    validate_commit_message(commit.message())?;

    info!("Advancing next to commit '{}'", commit);
    reset(
        open_repository,
        repo_details,
        &repo_config.branches().next(),
        &commit.into(),
        &force,
    )
    .map_err(|err| err.to_string())
    .map_err(AdvanceError::Unexpected)?;
    Ok(message_token)
}

#[instrument]
fn validate_commit_message(message: &Message) -> Result<(), AdvanceError> {
    let message = &message.to_string();
    if message.to_ascii_lowercase().starts_with("wip") {
        return Err(AdvanceError::NextCommitIsWIP);
    }
    match ::git_conventional::Commit::parse(message) {
        Ok(commit) => {
            info!(?commit, "Pass");
            Ok(())
        }
        Err(err) => {
            warn!(?err, "Fail");
            Err(AdvanceError::UnconventionalCommitMessage {
                reason: err.kind().to_string(),
            })
        }
    }
}

#[allow(clippy::cognitive_complexity)]
pub fn find_next_commit_on_dev(
    next: &Commit,
    main: &Commit,
    dev_commit_history: &[Commit],
) -> (Option<Commit>, Force) {
    let mut next_commit: Option<&Commit> = None;
    let mut force = Force::None;

    // Check if next is in dev branch's history
    let next_in_dev_history = dev_commit_history.iter().any(|commit| commit == next);

    // If next is not in dev history, we need to force push
    if !next_in_dev_history {
        debug!("Next commit not found in dev history, will use force push");
        force = Force::ForceWithLease {
            expected_sha: next.sha().to_string(),
        };
    }

    for commit in dev_commit_history {
        if commit == next {
            break;
        };
        if commit == main {
            debug!("Found main commit in dev history before next, will use force push");
            force = Force::ForceWithLease {
                expected_sha: next.sha().to_string(),
            };
            break;
        };
        next_commit.replace(commit);
    }

    debug!(?next_commit, ?force, "find_next_commit_on_dev result");
    (next_commit.cloned(), force)
}

// advance main branch to the commit 'next'
#[instrument(fields(next), skip_all)]
pub fn advance_main(
    next: Commit,
    repo_details: &RepoDetails,
    repo_config: &RepoConfig,
    open_repository: &dyn OpenRepositoryLike,
) -> Result<()> {
    info!("Advancing main to next");
    reset(
        open_repository,
        repo_details,
        &repo_config.branches().main(),
        &next.into(),
        &Force::None,
    )?;
    Ok(())
}
