use kameo::message::{Context, Message};
use tempfile::tempdir;

use crate::gitlib::Force;
use crate::prelude::*;
use crate::{
    core::{
        git::{Commit, GitRef},
        BranchName, GitDir, RepoConfig,
    },
    repo::{messages::RebaseDevOnMain, RepoActor},
    server::actor::messages::RepoUpdate,
};

impl Message<RebaseDevOnMain> for RepoActor {
    type Reply = Result<()>;

    async fn handle(
        &mut self,
        _msg: RebaseDevOnMain,
        _ctx: &mut Context<Self, Self::Reply>,
    ) -> Self::Reply {
        let Some(repo_config) = self.repo_details.repo_config.clone() else {
            self.alert_tui("no repo config").await?;
            self.post_repo_update(RepoUpdate::NoConfig).await?;
            info!("no repo config");
            return Ok(());
        };
        let Some(old_dev_commit) = &self.last_dev_commit else {
            info!("no old dev commit");
            return Ok(());
        };
        let Some(main_commit) = &self.last_main_commit else {
            info!("no old main commit");
            return Ok(());
        };

        info!("rebasing 'dev' onto 'main'");

        if let Err(err) = self
            .rebase_dev_on_main(&repo_config, old_dev_commit, main_commit)
            .await
        {
            self.post_repo_update(RepoUpdate::RebaseError(err.to_string()))
                .await
        } else {
            self.post_repo_update(RepoUpdate::RebaseCompleted).await
        }
    }
}
impl RepoActor {
    async fn rebase_dev_on_main(
        &self,
        repo_config: &RepoConfig,
        old_dev_commit: &Commit,
        main_commit: &Commit,
    ) -> Result<()> {
        let temp_dir = tempdir()?;

        let temp_gitdir = GitDir::new(temp_dir.path().to_path_buf());
        let temp_repo_details = self.repo_details.clone().with_gitdir(temp_gitdir);
        self.post_repo_update(RepoUpdate::ProgressMessage(s!("cloning repo")))
            .await?;
        let temp_repo = self.repository_factory.git_clone(&temp_repo_details)?;

        temp_repo.fetch()?;
        let dev = repo_config.branches().dev();
        let main = repo_config.branches().main();
        self.post_repo_update(RepoUpdate::ProgressMessage(s!("cloning repo")))
            .await?;
        temp_repo.checkout(&dev)?;
        self.post_repo_update(RepoUpdate::ProgressMessage(format!(
            "rebasing '{dev}' onto '{main}'"
        )))
        .await?;
        temp_repo.rebase(&BranchName::from(format!("origin/{main}")))?;
        self.post_repo_update(RepoUpdate::ProgressMessage(format!(
            "pushing rebased '{dev}' branch"
        )))
        .await?;
        let main_git_ref = GitRef::from(main_commit.clone());
        let old_dev_git_ref = GitRef::from(old_dev_commit.clone());
        temp_repo.push(
            &self.repo_details,
            &dev,
            &main_git_ref,
            &Force::ForceWithLease {
                expected_sha: old_dev_git_ref.to_string(),
            },
        )?;

        // temp_dir goes out of scope deleting the cloned repo
        Ok(())
    }
}
